# -*- coding: utf-8 -*-

# an alignment like
# TCA--CCCAT
# GGAGGTTGGT
# is represented in a list as ['TCA--CCCAT','GGAGGTTGGT']


# a similarity matrix is represented by a dict of dicts, so that the sililarity
# of base T in seq 1 and G in seq 2 can be referred as, e.g., sim['T']['G']

# some possible similarity matrices are reported hereafter

sim_m_0 = {'A': {'A':10, 'G':-1, 'C':-3, 'T':-4}, \
    'G': {'A':-1, 'G':7, 'C':-5, 'T':-3}, \
    'C': {'A':-3, 'G':-5, 'C':9, 'T':0}, \
    'T': {'A':-4, 'G':-3, 'C':0, 'T':8}}

sim_m_1 = {'A': {'A':10, 'G':-6, 'C':-6, 'T':-6}, \
    'G': {'A':-6, 'G':10, 'C':-6, 'T':-6}, \
    'C': {'A':-6, 'G':-6, 'C':10, 'T':-6}, \
    'T': {'A':-6, 'G':-6, 'C':-6, 'T':10}}

#the next is the one used in BLAST
sim_BLAST = {'A': {'A':5, 'G':-4, 'C':-4, 'T':-4}, \
    'G': {'A':-4, 'G':5, 'C':-4, 'T':-4}, \
    'C': {'A':-4, 'G':-4, 'C':5, 'T':-4}, \
    'T': {'A':-4, 'G':-4, 'C':-4, 'T':5}}
    
#the next is the one used in BLAST
sim_BLAST_RNA = {'A': {'A':5, 'G':-4, 'C':-4, 'U':-4}, \
    'G': {'A':-4, 'G':5, 'C':-4, 'U':-4}, \
    'C': {'A':-4, 'G':-4, 'C':5, 'U':-4}, \
    'U': {'A':-4, 'G':-4, 'C':-4, 'U':5}}

#the next has been proposed for aligning homologous RNA strands
sim_RIBOSUM85_60 = {'A': {'A':2.22, 'G':-1.46, 'C':-1.86, 'U':-1.39}, \
    'G': {'A':-1.46, 'G':1.03, 'C':-2.48, 'U':-1.74}, \
    'C': {'A':-1.86, 'G':-2.48, 'C':1.16, 'U':-1.05}, \
    'U': {'A':-1.39, 'G':-1.74, 'C':-1.05, 'U':1.65}}


#the next one considers transitions/transvections
sim_TR_TR = {'A': {'A':1, 'G':-1, 'C':-5, 'T':-5}, \
    'G': {'A':-1, 'G':1, 'C':-5, 'T':-5}, \
    'C': {'A':-5, 'G':-5, 'C':1, 'T':-1}, \
    'T': {'A':-5, 'G':-5, 'C':-1, 'T':1}}

#the next one corresponds to -1*(Levenshtein) (-edit_distance)
sim_neg_Leven = {'A': {'A':0, 'G':-1, 'C':-1, 'T':-1}, \
    'G': {'A':-1, 'G':0, 'C':-1, 'T':-1}, \
    'C': {'A':-1, 'G':-1, 'C':0, 'T':-1}, \
    'T': {'A':-1, 'G':-1, 'C':-1, 'T':0}}

PAM250 = {'A': {'A':  2, 'C': -2, 'B':  0, 'E':  0, 'D':  0, 'G':  1, 'F': -3, 'I': -1, 'H': -1, 'K': -1, 'M': -1, 'L': -2, 'N':  0, 'Q':  0, 'P':  1, 'S':  1, 'R': -2, 'T':  1, 'W': -6, 'V':  0, 'Y': -3, 'X':  0, 'Z':  0}, 
          'C': {'A': -2, 'C': 12, 'B': -4, 'E': -5, 'D': -5, 'G': -3, 'F': -4, 'I': -2, 'H': -3, 'K': -5, 'M': -5, 'L': -6, 'N': -4, 'Q': -5, 'P': -3, 'S':  0, 'R': -4, 'T': -2, 'W': -8, 'V': -2, 'Y':  0, 'X': -3, 'Z': -5}, 
          'B': {'A':  0, 'C': -4, 'B':  3, 'E':  3, 'D':  3, 'G':  0, 'F': -4, 'I': -2, 'H':  1, 'K':  1, 'M': -2, 'L': -3, 'N':  2, 'Q':  1, 'P': -1, 'S':  0, 'R': -1, 'T':  0, 'W': -5, 'V': -2, 'Y': -3, 'X': -1, 'Z':  2}, 
          'E': {'A':  0, 'C': -5, 'B':  3, 'E':  4, 'D':  3, 'G':  0, 'F': -5, 'I': -2, 'H':  1, 'K':  0, 'M': -2, 'L': -3, 'N':  1, 'Q':  2, 'P': -1, 'S':  0, 'R': -1, 'T':  0, 'W': -7, 'V': -2, 'Y': -4, 'X': -1, 'Z':  3}, 
          'D': {'A':  0, 'C': -5, 'B':  3, 'E':  3, 'D':  4, 'G':  1, 'F': -6, 'I': -2, 'H':  1, 'K':  0, 'M': -3, 'L': -4, 'N':  2, 'Q':  2, 'P': -1, 'S':  0, 'R': -1, 'T':  0, 'W': -7, 'V': -2, 'Y': -4, 'X': -1, 'Z':  3}, 
          'G': {'A':  1, 'C': -3, 'B':  0, 'E':  0, 'D':  1, 'G':  5, 'F': -5, 'I': -3, 'H': -2, 'K': -2, 'M': -3, 'L': -4, 'N':  0, 'Q': -1, 'P':  0, 'S':  1, 'R': -3, 'T':  0, 'W': -7, 'V': -1, 'Y': -5, 'X': -1, 'Z':  0}, 
          'F': {'A': -3, 'C': -4, 'B': -4, 'E': -5, 'D': -6, 'G': -5, 'F':  9, 'I':  1, 'H': -2, 'K': -5, 'M':  0, 'L':  2, 'N': -3, 'Q': -5, 'P': -5, 'S': -3, 'R': -4, 'T': -3, 'W':  0, 'V': -1, 'Y':  7, 'X': -2, 'Z': -5}, 
          'I': {'A': -1, 'C': -2, 'B': -2, 'E': -2, 'D': -2, 'G': -3, 'F':  1, 'I':  5, 'H': -2, 'K': -2, 'M':  2, 'L':  2, 'N': -2, 'Q': -2, 'P': -2, 'S': -1, 'R': -2, 'T':  0, 'W': -5, 'V':  4, 'Y': -1, 'X': -1, 'Z': -2}, 
          'H': {'A': -1, 'C': -3, 'B':  1, 'E':  1, 'D':  1, 'G': -2, 'F': -2, 'I': -2, 'H':  6, 'K':  0, 'M': -2, 'L': -2, 'N':  2, 'Q':  3, 'P':  0, 'S': -1, 'R':  2, 'T': -1, 'W': -3, 'V': -2, 'Y':  0, 'X': -1, 'Z':  2}, 
          'K': {'A': -1, 'C': -5, 'B':  1, 'E':  0, 'D':  0, 'G': -2, 'F': -5, 'I': -2, 'H':  0, 'K':  5, 'M':  0, 'L': -3, 'N':  1, 'Q':  1, 'P': -1, 'S':  0, 'R':  3, 'T':  0, 'W': -3, 'V': -2, 'Y': -4, 'X': -1, 'Z':  0}, 
          'M': {'A': -1, 'C': -5, 'B': -2, 'E': -2, 'D': -3, 'G': -3, 'F':  0, 'I':  2, 'H': -2, 'K':  0, 'M':  6, 'L':  4, 'N': -2, 'Q': -1, 'P': -2, 'S': -2, 'R':  0, 'T': -1, 'W': -4, 'V':  2, 'Y': -2, 'X': -1, 'Z': -2}, 
          'L': {'A': -2, 'C': -6, 'B': -3, 'E': -3, 'D': -4, 'G': -4, 'F':  2, 'I':  2, 'H': -2, 'K': -3, 'M':  4, 'L':  6, 'N': -3, 'Q': -2, 'P': -3, 'S': -3, 'R': -3, 'T': -2, 'W': -2, 'V':  2, 'Y': -1, 'X': -1, 'Z': -3}, 
          'N': {'A':  0, 'C': -4, 'B':  2, 'E':  1, 'D':  2, 'G':  0, 'F': -3, 'I': -2, 'H':  2, 'K':  1, 'M': -2, 'L': -3, 'N':  2, 'Q':  1, 'P':  0, 'S':  1, 'R':  0, 'T':  0, 'W': -4, 'V': -2, 'Y': -2, 'X':  0, 'Z':  1}, 
          'Q': {'A':  0, 'C': -5, 'B':  1, 'E':  2, 'D':  2, 'G': -1, 'F': -5, 'I': -2, 'H':  3, 'K':  1, 'M': -1, 'L': -2, 'N':  1, 'Q':  4, 'P':  0, 'S': -1, 'R':  1, 'T': -1, 'W': -5, 'V': -2, 'Y': -4, 'X': -1, 'Z':  3}, 
          'P': {'A':  1, 'C': -3, 'B': -1, 'E': -1, 'D': -1, 'G':  0, 'F': -5, 'I': -2, 'H':  0, 'K': -1, 'M': -2, 'L': -3, 'N':  0, 'Q':  0, 'P':  6, 'S':  1, 'R':  0, 'T':  0, 'W': -6, 'V': -1, 'Y': -5, 'X': -1, 'Z':  0}, 
          'S': {'A':  1, 'C':  0, 'B':  0, 'E':  0, 'D':  0, 'G':  1, 'F': -3, 'I': -1, 'H': -1, 'K':  0, 'M': -2, 'L': -3, 'N':  1, 'Q': -1, 'P':  1, 'S':  2, 'R':  0, 'T':  1, 'W': -2, 'V': -1, 'Y': -3, 'X':  0, 'Z':  0}, 
          'R': {'A': -2, 'C': -4, 'B': -1, 'E': -1, 'D': -1, 'G': -3, 'F': -4, 'I': -2, 'H':  2, 'K':  3, 'M':  0, 'L': -3, 'N':  0, 'Q':  1, 'P':  0, 'S':  0, 'R':  6, 'T': -1, 'W':  2, 'V': -2, 'Y': -4, 'X': -1, 'Z':  0}, 
          'T': {'A':  1, 'C': -2, 'B':  0, 'E':  0, 'D':  0, 'G':  0, 'F': -3, 'I':  0, 'H': -1, 'K':  0, 'M': -1, 'L': -2, 'N':  0, 'Q': -1, 'P':  0, 'S':  1, 'R': -1, 'T':  3, 'W': -5, 'V':  0, 'Y': -3, 'X':  0, 'Z': -1}, 
          'W': {'A': -6, 'C': -8, 'B': -5, 'E': -7, 'D': -7, 'G': -7, 'F':  0, 'I': -5, 'H': -3, 'K': -3, 'M': -4, 'L': -2, 'N': -4, 'Q': -5, 'P': -6, 'S': -2, 'R':  2, 'T': -5, 'W': 17, 'V': -6, 'Y':  0, 'X': -4, 'Z': -6}, 
          'V': {'A':  0, 'C': -2, 'B': -2, 'E': -2, 'D': -2, 'G': -1, 'F': -1, 'I':  4, 'H': -2, 'K': -2, 'M':  2, 'L':  2, 'N': -2, 'Q': -2, 'P': -1, 'S': -1, 'R': -2, 'T':  0, 'W': -6, 'V':  4, 'Y': -2, 'X': -1, 'Z': -2}, 
          'Y': {'A': -3, 'C':  0, 'B': -3, 'E': -4, 'D': -4, 'G': -5, 'F':  7, 'I': -1, 'H':  0, 'K': -4, 'M': -2, 'L': -1, 'N': -2, 'Q': -4, 'P': -5, 'S': -3, 'R': -4, 'T': -3, 'W':  0, 'V': -2, 'Y': 10, 'X': -2, 'Z': -4}, 
          'X': {'A':  0, 'C': -3, 'B': -1, 'E': -1, 'D': -1, 'G': -1, 'F': -2, 'I': -1, 'H': -1, 'K': -1, 'M': -1, 'L': -1, 'N':  0, 'Q': -1, 'P': -1, 'S':  0, 'R': -1, 'T':  0, 'W': -4, 'V': -1, 'Y': -2, 'X': -1, 'Z': -1}, 
          'Z': {'A':  0, 'C': -5, 'B':  2, 'E':  3, 'D':  3, 'G':  0, 'F': -5, 'I': -2, 'H':  2, 'K':  0, 'M': -2, 'L': -3, 'N':  1, 'Q':  3, 'P':  0, 'S':  0, 'R':  0, 'T': -1, 'W': -6, 'V': -2, 'Y': -4, 'X': -1, 'Z':  3}}


def print_alignment(al):
    """
    A nice, standard textual output for an alignment
    """
    string_upper=al[0]
    string_lower=al[1]
    string_in_between=''
    for sx,sy in zip(al[0],al[1]):
        if sx == sy:
            string_in_between+="|"
        else:
            string_in_between+=" "
    print "  ",string_upper
    print "  ",string_in_between
    print "  ",string_lower



def compute_matrix_F(a, b, scoring_matrix, gap_penalty):
    """
    Returns the score matrix F for two sequences a and b to be aligned.
    F is the matrix for partial scores.
    The matrix is represented as a list of lists, with #columns =len(a)+1
    and #rows = len(b)+1
    """

    len_a = len(a)
    len_b = len(b)

    matrix_F = []

    for i in xrange(len_b + 1):
        temp = []
        for j in xrange(len_a + 1):
            if i == 0:
                temp.append(gap_penalty * j)
            elif j == 0:
                temp.append(gap_penalty * i)
            else:
                temp.append(0)

        matrix_F.append(temp)

    #computation of scores

    for i in xrange(1, len_b + 1):
        for j in xrange(1, len_a + 1):
            val1 = matrix_F[i-1][j-1] + scoring_matrix[a[j-1]][b[i-1]]
            val2 = matrix_F[i][j-1] + gap_penalty
            val3 = matrix_F[i-1][j] + gap_penalty
            val = max(val1, val2, val3)
            matrix_F[i][j] = val

    return matrix_F



def g_align_nw(a, b, sc_matrix=sim_BLAST, gap_penalty=-1.6):
    """
    global alignment with Needleman Wunsch;
    returns a list with all the alignments with max score
    """

    matrix_F = compute_matrix_F(a, b, sc_matrix, gap_penalty)

    all_alignments = []

    def find_alignments(all_alignments, matrix_F, index_a, index_b, partial_a, partial_b):
        
        i = index_b
        j = index_a

        if(i > 0 and j > 0):
            score = matrix_F[i][j]
            
            score_diag = matrix_F[i - 1][j - 1]
            score_up = matrix_F[i-1][j]
            score_left = matrix_F[i][j-1]

            if score == score_diag + sc_matrix[a[j-1]][b[i-1]]:
                new_partial_a = a[j-1] + partial_a
                new_partial_b = b[i-1] + partial_b
                find_alignments(all_alignments, matrix_F, j-1, i-1, new_partial_a[:], new_partial_b[:])

            if score == score_left + gap_penalty:
                new_partial_a = a[j-1] + partial_a
                new_partial_b = '-' + partial_b
                find_alignments(all_alignments, matrix_F, j-1, i, new_partial_a[:], new_partial_b[:])

            if score == score_up + gap_penalty:
                new_partial_a = '-' + partial_a
                new_partial_b = b[i-1] + partial_b
                find_alignments(all_alignments, matrix_F, j, i-1, new_partial_a[:], new_partial_b[:])
        else:

            while j > 0:
                partial_a = a[j-1] + partial_a
                partial_b = '-' + partial_b
                j -= 1
            while i > 0:
                partial_a = '-' + partial_a
                partial_b = b[i-1] + partial_b
                i -= 1

            current_alignment = []
            current_alignment.append(partial_a)
            current_alignment.append(partial_b)

            all_alignments.append(current_alignment)

        return all_alignments

    return find_alignments(all_alignments, matrix_F, len(matrix_F[0])-1, len(matrix_F)-1, '', '')



def g_align_nw_simple(a, b, sc_matrix, gap_penalty):


    matrix_F = compute_matrix_F(a,b,sc_matrix, gap_penalty)

    al_upper_seq = ''
    al_lower_seq = ''
    i = len(b)
    j = len(a)
    while i > 0 and j > 0:
        score = matrix_F[i][j]
        score_diag = matrix_F[i - 1][j - 1]
        score_up = matrix_F[i-1][j]
        score_left = matrix_F[i][j-1]
        if score == score_diag + sc_matrix[a[j-1]][b[i-1]]:
            al_upper_seq = a[j-1] + al_upper_seq
            al_lower_seq = b[i-1] + al_lower_seq
            i -= 1
            j -= 1
        elif score == score_left + gap_penalty:
            al_upper_seq = a[j-1] + al_upper_seq
            al_lower_seq = '-' + al_lower_seq
            j -= 1
        elif score == score_up + gap_penalty:
            al_upper_seq = '-' + al_upper_seq
            al_lower_seq = b[i-1] + al_lower_seq
            i -= 1

    #the following is done to go to [0,0] once the first row/column is reached
    while j > 0:
        al_upper_seq = a[j-1] + al_upper_seq
        al_lower_seq = '-' + al_lower_seq
        j -= 1
    while i > 0:
        al_upper_seq = '-' + al_upper_seq
        al_lower_seq = b[i-1] + al_lower_seq
        i -= 1

    return [al_upper_seq, al_lower_seq]