# -*- coding: utf-8 -*-

#4a) Si consideri un insieme di N specie. Per ciascuna specie è conosciuta
# la sequenza di amminoacidi di una data proteina omologa (o parte di essa). 
# Una distanza tra sequenze omologhe di due specie diverse potrebbe essere 
# il numero di gap presenti nell'allineamento; chiamiamola "gap_distance".
# 
# Si sviluppi una funzione "gapdist_mat(file_list)" che costruisca la matrice 
# delle gap_distance tra sequenze omologhe di N specie diverse, a partire da  
# una lista di nomi di file fasta con i relativi allineamenti a coppie.
# A ogni specie è associato un indice univoco, con valore compreso tra 0 e N-1.
# Inoltre, il nome di un file deve seguire la seguente convenzione: 
# "alignment03.fasta" indica l'allineamento tra la specie "0" e quella "3". 
# La dimensione della matrice in output (nella forma "lista di liste") 
# dipenderà dunque dal massimo indice presente nei nomi dei file in ingresso.
# Nota: qualora non si abbia a disposizione in input l'allineamento tra X e Y, 
# nelle due celle corrispondenti dovrà essere presente il valore None.
#
# Es: 
# l = ["alignment01.fasta","alignment02.fasta","alignment12.fasta"]
# gapdist_mat(l)  --> [[ 0, 34, 49]
#                      [34,  0, 46]
#                      [49, 46,  0]]
# l = ["alignment12.fasta","alignment23.fasta","alignment02.fasta"]
# gapdist_mat(l)  --> [[   0, None, 49, None]
#                      [None,    0, 46, None]
#                      [  49,   46,  0,   48]
#                      [None, None, 48,    0]]


import Bio.SeqIO

from soluzione2 import num_gaps

def gapdist_mat(file_list):
    
    #funct to obtain a list with the 2 indexes out of the filename
    def getindices(filename):
        indexchars = \
        list(filename.replace(".fasta","").replace("alignment",""))
        return sorted(int(x) for x in indexchars)
    
    #let's obtain a list with all the indices from all the filenames
    indices = []
    for name in file_list:
        for x in getindices(name):
            indices.append(x)        
    mi = max(indices) #max index, so to initialize the result matrix
    
    #matrix initialization
    res = [[None if i!=j else 0 for j in xrange(mi+1)] for i in xrange(mi+1)] 
    
    #insertion of each available entry
    for filename in file_list:
        with open(filename) as f_in:
            al = list(str(s.seq) for s in Bio.SeqIO.parse(f_in, "fasta"))
        n = num_gaps(al)
        x, y = getindices(filename)
        res[x][y] = res[y][x] = n
        
    return res

